@@ -10,6 +10,13 @@ from .b3cube import plot_richness
 from .b3cube import cumulative_species
 from .b3cube import calculate_rate
 from .b3cube import get_survey_effort
+from .b3cube import plot_cumsum
+from .b3cube import filter_multiple_cells
+from .b3cube import filter_multiple_occ
+from .b3cube import aggregate_count_per_cell
+from .b3cube import filter_time_window
+from .b3cube import find_correlations
+
 
 __all__ = [
     "plot_richness",
@@ -18,6 +25,9 @@ __all__ = [
     "filter_multiple_cells",
     "filter_multiple_occ",
     "calculate_rate",
-    "get_survey_effort"
+    "get_survey_effort",
+    "aggregate_count_per_cell",
+    "filter_time_window",
+    "find_correlations"
     # purposely exclude OccurrenceCube
 ]
\ No newline at end of file
