@@ -1,21 +1,26 @@
+import os
 import geopandas as gpd
 import pandas as pd
 import xarray as xr
+import numpy as np
 import sparse
 import dask.array as da
-import numpy as np
 import matplotlib.pyplot as plt
 import matplotlib
-matplotlib.use("Agg")
-import shapely
-from shapely import geos
 import gcsfs
+import folium
+from folium import Choropleth
+from IPython.display import display
+from b3alien.utils.runtime import in_jupyter
+matplotlib.use("Agg")
 
 """
 
 Create a class object of the occurrence cube
 
 """
+
+
 class OccurrenceCube():
 
     """
@@ -31,7 +36,6 @@ class OccurrenceCube():
             Optional coordinates to assign to the cube.
         index_col : str or list, optional
             Column(s) to use for reshaping if needed.
-            
 
         Returns
         -------
@@ -41,25 +45,37 @@ class OccurrenceCube():
             self.data a sparse xarray.Xarray
     """
 
-    def __init__(self, filepath: str, dims=None, coords=None, index_col=None):
-       
+    def __init__(self, filepath: str, source='geoparquet', gproject='', dims=None, coords=None, index_col=None):
+
         self.filepath = filepath
+        self.source = source
+        self.gproject = gproject
         self.dims = dims or ("time", "cell", "species")
         self.coords = coords
         self.index_col = index_col
 
-        # Load GeoParquet
-        self.df = self._load_geoparquet(filepath)
-        
+        # Load Data
+        if source == 'gbif':
+            self.df = self._load_gbifcsv(filepath)
+        else:
+            self.df = self._load_geoparquet(filepath, gproject)
+
         # Create cube
         self.data = self._create_xcube(self.df)
 
-    def _load_geoparquet(self, path):
+    def _load_gbifcsv(self, path):
         """
-        Load a GeoParquet file from local disk or GCS using GeoPandas.
+        Load a GBIF CSV file from local disk using Pandas. Assumes tab-separated values.
+        """
+        df = pd.read_csv(path, sep='\t')
+        return df
+
+    def _load_geoparquet(self, path, gproject):
+        """
+        Load a GeoParquet file from local disk or GCS using GeoPandas. 
         """
         if path.startswith("gs://"):
-            fs = gcsfs.GCSFileSystem()
+            fs = gcsfs.GCSFileSystem(project=gproject)
             with fs.open(path) as f:
                 gdf = gpd.read_parquet(f)
         else:
@@ -74,19 +90,16 @@ class OccurrenceCube():
 
         return gdf
 
-    def _create_xcube(self, df):
+    def _create_xcube(self, df, dims=("time", "cell", "species")):
         """
-        Convert a GeoDataFrame into a sparse xarray cube with geometry metadata.
+        This function converts a GeoDataFrame into a sparse xarray cube with geometry metadata in case a GeoParquet file was loaded.
+        In case of a pure GBIF cube, the geometry is ignored.  
         """
         # Convert to categorical
         df["yearmonth"] = pd.Categorical(df["yearmonth"])
         df["cellCode"] = pd.Categorical(df["cellCode"])
         df["specieskey"] = pd.Categorical(df["specieskey"])
 
-        # Align geometries with cell categories
-        cell_categories = df["cellCode"].cat.categories
-        geometry_per_cell = df.drop_duplicates("cellCode").set_index("cellCode").loc[cell_categories]["geometry"]
-
         # Encode to integers
         time_codes = df["yearmonth"].cat.codes.values
         cell_codes = df["cellCode"].cat.codes.values
@@ -103,22 +116,41 @@ class OccurrenceCube():
             )
         )
 
+        # Prepare coordinates for the DataArray
+        coords = {
+            dims[0]: df["yearmonth"].cat.categories,
+            dims[1]: df["cellCode"].cat.categories,
+            dims[2]: df["specieskey"].cat.categories,
+        }
+
+        # Conditionally add geometry if the column exists
+        if "geometry" in df.columns:
+            cell_categories = df["cellCode"].cat.categories
+            geometry_per_cell = df.drop_duplicates("cellCode").set_index("cellCode").loc[cell_categories]["geometry"]
+            coords["geometry"] = (dims[1], geometry_per_cell.values)
+
         # Create xarray DataArray
         cube = xr.DataArray(
             sparse_cube,
             dims=self.dims,
-            coords={
-                self.dims[0]: df["yearmonth"].cat.categories,
-                self.dims[1]: df["cellCode"].cat.categories,
-                self.dims[2]: df["specieskey"].cat.categories,
-                "geometry": (self.dims[1], geometry_per_cell.values)
-            },
+            coords=coords,
             name="occurrences"
         )
 
         return cube
 
     def _species_richness(self, normalized=False):
+        """
+            Calculate species richness per cell from the sparse cube.
+            Parameters
+            ----------
+            normalized : bool
+                Whether to compute normalized richness (richness / total occurrences).
+            Returns
+            -------
+            None
+                Sets self.richness as a DataFrame with columns 'cell' and 'richness' or 'normalized_richness'.
+        """
         # 1. Binary presence
         presence = (self.data > 0)
 
@@ -130,7 +162,7 @@ class OccurrenceCube():
 
         total_occurrences = self.data.sum(dim=["time", "species"])
 
-        if normalized == False:
+        if not normalized:
             # 4. Get the non-zero values and indices
             coords = species_richness.data.coords  # (1D arrays of indices)
             values = species_richness.data.data    # the richness values
@@ -162,45 +194,216 @@ class OccurrenceCube():
             self.richness = norm_df
 
     def _filter_species(self, speciesKey):
+        """
+            Filter the cube to only include data for a specific speciesKey.
+
+            Parameters
+            ----------
+            speciesKey : int or str
+                The GBIF speciesKey to filter on.
+
+            Returns
+            -------
+            None
+                Updates self.df and self.data to only include the specified speciesKey.
+        """
 
         self.df = self.df[self.df['specieskey'].eq(speciesKey)]
         self.data = self.data.sel(species=speciesKey)
 
-def plot_richness(richness_df, gdf_from_gcs, geom='cellCode'):
+
+def find_correlations(cube: OccurrenceCube, top_n=10):
     """
-        Create a plot of the species richness dataframe.
+    Calculates species co-occurrence using efficient matrix operations.
 
-        Parameters
-        ----------
-        richness_df : pandas.DataFrame
-            Datagrame containing the species richness per grid cell.
-        gdf_from_gcs : geopandas.Dataframe
-            GeoDataFrame containing the species occurrence cuve.
-        geom : str, optional
-            Name of the geometry column in the GeoDataFrame. Default is 'cellCode'
-            
-        Returns
-        -------
-        matplotlib.plot
-            A plot of the species richness.
+    This method identifies pairs of species that occur in the same 'cell'
+    at the same 'time'.
+
+    Parameters
+    ----------
+        cube (OccurrenceCube): An instance of the OccurrenceCube class.
+        top_n (int): The number of top correlated pairs to return.
+
+    Returns
+    -------
+        list: A list of tuples, where each tuple contains the species names
+              and their co-occurrence count, sorted in descending order.
     """
+    if cube.data is None:
+        print("Data cube is not loaded. Cannot find correlations.")
+        return []
+        
+    # 1. Create a boolean cube for presence (occurrence > 0)
+    # Using int32 to prevent overflow when summing large arrays.
+    presence = (cube.data > 0).astype('int32')
 
-    gdf_plot = pd.merge(richness_df, gdf_from_gcs, left_on='cell', right_on=geom)
+    # 2. Compute the co-occurrence matrix using broadcasting.
+    # This multiplies the presence array with itself along a new 'species_copy' dimension
+    # and sums over the 'time' and 'cell' dimensions.
+    # The result is a (species x species) matrix.
+    co_occurrence = (presence * presence.rename({'species': 'species_copy'})).sum(dim=['time', 'cell'])
 
-    gdf_plot = gpd.GeoDataFrame(gdf_plot, geometry="geometry", crs=gdf_from_gcs.crs)
+    # 3. If the data is a Dask array, trigger the computation.
+    # Then, explicitly densify the sparse array before converting to pandas.
+    computed_co = co_occurrence.compute()
+    df_co = xr.DataArray(
+        computed_co.data.todense(),
+        dims=computed_co.dims,
+        coords=computed_co.coords
+    ).to_pandas()
+    df_co.columns.name = 'species'
+    df_co.index.name = 'species'
 
-    fig, ax = plt.subplots(figsize=(10, 10))
-    gdf_plot.plot(
-        column="richness",
-        cmap="viridis",
-        legend=True,
-        linewidth=0.1,
-        edgecolor="grey",
-        ax=ax
+    # 4. Extract the upper triangle to avoid duplicate pairs (A,B vs B,A)
+    # and self-pairs (A,A). k=1 excludes the diagonal.
+    df_upper = df_co.where(np.triu(np.ones(df_co.shape), k=1).astype(bool))
+    
+    # 5. Stack the matrix to get a Series of (species, species) -> count
+    stacked_pairs = df_upper.stack().sort_values(ascending=False)
+
+    # Filter out pairs with zero co-occurrences
+    strongest_pairs = stacked_pairs[stacked_pairs > 0]
+    
+    if strongest_pairs.empty:
+        print("No co-occurring species pairs found in the provided dataset.")
+        return []
+
+    # 6. Create a mapping from specieskey to species name for readability
+    species_map = cube.df.drop_duplicates('specieskey').set_index('specieskey')['species']
+    
+    # 7. Format the output
+    results = []
+    for (key1, key2), count in strongest_pairs.head(top_n).items():
+        name1 = species_map.get(key1, f"Unknown ({key1})")
+        name2 = species_map.get(key2, f"Unknown ({key2})")
+        results.append(((name1, name2), int(count)))
+        
+    return results
+
+
+def filter_time_window(df, start_year, end_year, cols=["year", "rate"]):
+    """Filter a two column dataframe based on year window.
+    
+    Parameters
+    ----------
+    df: pandas.DataFrame
+        A two column dataframe with one time axs
+    start_year : np.int
+        Integer number of the earliest date to filter on.
+    end_year : np.int
+        Integer number of the latest date to filter on.
+    cols: list of str, optional
+        List of the names of the columns of the dataframe
+    
+
+    Returns
+    -------
+    time: pandas.Series
+        Filtered time series
+    filtered_var: pandas.Series
+        Filtered variable corresponding to the time series
+    
+    """
+    filtered = df[(df[cols[0]] >= start_year) & (df[cols[0]] <= end_year)].reset_index(drop=True)
+    return filtered[cols[0]], filtered[cols[1]]
+
+
+def aggregate_count_per_cell(cube, taxonRank, taxon):
+    """
+    Aggregate the counts per taxonomic level per cell. This can be used as a
+    normalization factor
+
+    Parameters
+    ----------
+    cube: OccurrenceCube
+        An OccurrenceCube with geometries added
+    taxonRank: dwc.taxonRank term
+        Level at which the aggregation needs to be performed
+    taxon: str
+        Name of the taxon at which aggregation needs to be performed
+    plot: bool, optional
+        whether the aggregated count per cell needs to be plotted on a map
+
+    Returns
+    -------
+    gdf: GeoDataFrame
+        dataframe containing geometries and the aggregated occurrence count
+
+    """
+    # Step 1: Filter by the desired class
+    filtered_df = cube.df[cube.df[taxonRank] == taxon]
+
+    # Step 2: Select only the relevant columns
+    count_str = taxonRank + "count"
+    columns_to_keep = ["yearmonth", "cellCode", taxonRank, count_str, "geometry"]
+    subset_df = filtered_df[columns_to_keep]
+
+    # Step 3: Drop duplicate rows based on those columns
+    gdf = subset_df.drop_duplicates()
+
+    return gdf
+
+
+def plot_richness(cube, normalized=False, html_path='richness_map.html'):
+    """
+    Plot species richness from an OccurrenceCube instance.
+
+    Parameters
+    ----------
+    cube : OccurrenceCube
+        An instance of the cube with `.data` and `.richness` available.
+    normalized : bool
+        Whether to use normalized richness.
+    html_path : str
+        Path to save the map if not in Jupyter.
+
+    Returns
+    -------
+    None
+    """
+    # Compute richness if needed
+    if not hasattr(cube, "richness"):
+        cube._species_richness(normalized=normalized)
+
+    # Extract richness and geometry directly from cube
+    richness_df = cube.richness
+    geometry = cube.data.coords["geometry"].values
+    cell_labels = cube.data.coords["cell"].values
+
+    # Create GeoDataFrame
+    gdf_plot = gpd.GeoDataFrame(
+        richness_df,
+        geometry=geometry[richness_df["cell"].map(dict(zip(cell_labels, range(len(cell_labels)))))]
     )
-    ax.set_title("Species Richness per QDGC Cell")
-    ax.axis("off")
-    plt.show()
+    gdf_plot.set_crs("EPSG:4326", inplace=True)  # ensure WGS84 for folium
+
+    # Optionally simplify geometry
+    gdf_plot["geometry"] = gdf_plot["geometry"].simplify(0.01)
+
+    # Center map
+    centroid = gdf_plot.geometry.union_all().centroid
+    m = folium.Map(location=[centroid.y, centroid.x], zoom_start=7)
+
+    # Plot
+    Choropleth(
+        geo_data=gdf_plot,
+        name='Species Richness',
+        data=gdf_plot,
+        columns=['cell', 'richness' if not normalized else 'normalized_richness'],
+        key_on='feature.properties.cell',
+        fill_color='YlGnBu',
+        fill_opacity=0.7,
+        line_opacity=0.2,
+        legend_name='Species Richness'
+    ).add_to(m)
+
+    folium.LayerControl().add_to(m)
+
+    if in_jupyter():
+        display(m)
+    else:
+        m.save(html_path)
+        print(f"Map saved as HTML to {os.path.abspath(html_path)}")
 
 
 def cumulative_species(cube, species_to_keep):
@@ -215,7 +418,7 @@ def cumulative_species(cube, species_to_keep):
         species_to_keep : numpy.array
             Array of GBIF speciesKeys that need to be taken into account to calculate the cumulative species number of a cube.
         geom : str, optional
-            
+
         Returns
         -------
         df1 : pandas.DataFrame
@@ -223,7 +426,7 @@ def cumulative_species(cube, species_to_keep):
         df2 : pandas.DataFrame
             Cumulative dataframe cell independent.
     """
-    
+
     # Wrap sparse array in Dask array with one or more chunks
     dask_sparse_array = da.from_array(cube.data.data, chunks=(100, 100, 1000))  # tune chunking for your use case
 
@@ -287,6 +490,7 @@ def cumulative_species(cube, species_to_keep):
 
     return df_sparse, df_cumulative
 
+
 def plot_cumsum(df_cumulative):
     """
         Create a plot of the cumulative number of species.
@@ -295,13 +499,13 @@ def plot_cumsum(df_cumulative):
         ----------
         df_cumulative : pandas.DataFrame
             Datagrame containing the cumulative number over tume.
-            
+
         Returns
         -------
         matplotlib.plot
             A plot of the cumulative number of species.
     """
-    
+
     plt.figure(figsize=(10, 5))
     plt.plot(df_cumulative["time"], df_cumulative["cumulative_species"], marker="o")
     plt.title("Cumulative Unique Species Over Time")
@@ -321,7 +525,7 @@ def filter_multiple_cells(df_sparse):
         ----------
         df_sparse : pandas.DataFrame
             Datagrame containing the species richness per grid cell.
-            
+
         Returns
         -------
         pandas.DataFrame
@@ -351,7 +555,9 @@ def filter_multiple_cells(df_sparse):
         seen_species.update(new_species)
         cumulative.append((time, len(seen_species)))
 
-    df_cumulative_cells = pd.DataFrame(cumulative, columns=["time", "cumulative_species_2cells"])
+    df_cumulative_cells = pd.DataFrame(cumulative,
+                                       columns=["time",
+                                                "cumulative_species_2cells"])
 
     return df_cumulative_cells
 
@@ -363,7 +569,7 @@ def filter_multiple_occ(df_sparse):
         ----------
         df_sparse : pandas.DataFrame
             Datagrame containing the species richness per grid cell.
-            
+
         Returns
         -------
         pandas.DataFrame
@@ -396,19 +602,21 @@ def filter_multiple_occ(df_sparse):
         seen_species.update(new_species)
         cumulative.append((time, len(seen_species)))
 
-    df_cumulative_occ = pd.DataFrame(cumulative, columns=["time", "cumulative_species_2occ"])
+    df_cumulative_occ = pd.DataFrame(cumulative, 
+                                     columns=["time",
+                                              "cumulative_species_2occ"])
 
     return df_cumulative_occ
 
 def calculate_rate(df_cumulative):
     """
-       Calculate the rate of establishment from the cumulative distribution.
+        Calculate the rate of establishment from the cumulative distribution.
 
         Parameters
         ----------
         df_cumulative : pandas.DataFrame
             Datagrame containing the cumulative distribution.
-            
+
         Returns
         -------
         s1 : pandas.Series
@@ -465,14 +673,12 @@ def get_survey_effort(cube, dateFormat='%Y-%m', calc_type='total'):
                 'distinct' : total number of distinct observers
                 'total' : total number of occurrences
                 Default is total.
-            
+
         Returns
         -------
         df : pandas.DataFrame
             Dataframe containing time and the chosen measurement for survey effort.
     """
-    
-
     if calc_type == 'distinct':
         # Group by 'yearmonth' and sum
         distinct_observers_over_time = cube.df.groupby('yearmonth', observed=True)['distinctobservers'].sum()
@@ -508,4 +714,5 @@ def get_survey_effort(cube, dateFormat='%Y-%m', calc_type='total'):
         # (Optional) Drop rows with invalid or missing time
         df = df.dropna(subset=['time'])
 
-        return df
\ No newline at end of file
+        return df
+
