@@ -10,5 +10,10 @@ from .griis import get_speciesKey
 from .griis import do_taxon_matching
 from .griis import read_checklist
 from .griis import split_event_date
+from .griis import get_species_under_genus
 
-__all__ = ["get_speciesKey", "do_taxon_matching", "read_checklist", "split_event_date"]
\ No newline at end of file
+__all__ = ["get_speciesKey",
+           "do_taxon_matching",
+           "read_checklist",
+           "split_event_date",
+           "get_species_under_genus"]
